// Editor related extensions, not available on Linux

#include "XC_Engine.h"
#include "UnPath.h"

IMPLEMENT_CLASS(AXC_PathBuilder);


//#define DEBUGGINGPATHS  1 //1 to put path info in log
#define MAXMARKERS 3000 //bound number of turn markers
#define MAXREACHSPECS 3000 //bound number of reachspecs 
#define MAXCOMMONRADIUS 70 //max radius to consider in building paths
#define MAXCOMMONHEIGHT 70
#define MINCOMMONHEIGHT 48 //min typical height for non-human intelligent creatures
#define MINCOMMONRADIUS 24 //min typical radius for non-human intelligent creatures
#define COMMONRADIUS    52 //max typical radius of intelligent creatures

enum EReachSpecFlags
{
	R_WALK = 1,	//walking required
	R_FLY = 2,   //flying required 
	R_SWIM = 4,  //swimming required
	R_JUMP = 8,   // jumping required
	R_DOOR = 16,
	R_SPECIAL = 32,
	R_PLAYERONLY = 64
}; 

static float HUMANRADIUS = 17.f;
static float HUMANHEIGHT = 39.f;
static UClass* InventorySpotClass = NULL;
static float ScanRangeSq = 1000000.f;

void AXC_PathBuilder::StaticConstructor()
{
	UClass* MyClass = GetClass();
	if ( GIsEditor )
		MyClass->ClassFlags &= ~CLASS_Transient;

	AXC_PathBuilder* DefaultObject = GetDefault<AXC_PathBuilder>();
	DefaultObject->MaxScanRange = 1000.f;
	DefaultObject->InventorySpotType = AInventorySpot::StaticClass();

	new(MyClass, TEXT("MaxScanRange"), RF_Public) UFloatProperty(CPP_PROPERTY(MaxScanRange), TEXT("PathBuilder"), CPF_Native|CPF_Edit);
	new(MyClass, TEXT("InventorySpotClass"), RF_Public) UClassProperty(CPP_PROPERTY(InventorySpotType), TEXT("PathBuilder"), CPF_Native|CPF_Edit, AInventorySpot::StaticClass() );
}

void AXC_PathBuilder::InitExecution()
{
	Super::InitExecution();

	if ( GIsEditor )
	{
		InventorySpotClass = InventorySpotType;
		ScanRangeSq = MaxScanRange * MaxScanRange;
		Builder.undefinePaths_XC( GetLevel() );
		Builder.definePaths_XC( GetLevel() );
	}
	GetLevel()->DestroyActor( this);
}


//====================================================
//====================================================
// Enhanced path builder
//====================================================
//====================================================

struct ActorIdxPair
{
	AActor* Actor;
	INT Idx;
	
	ActorIdxPair( AActor* InActor, INT InIdx)
	: Actor(InActor)
	, Idx(InIdx)
	{}
};

void XC_ENGINE_API FXC_PathBuilder::definePaths_XC (ULevel *ownerLevel)
{
	guard(FXC_PathBuilder::definePaths_XC);

/*	TArray<ActorIdxPair> Pairs(0);
	//Add TriggerMarkers to unpathed triggers
	
	INT i;
	for (i=0; i<Level->Actors.Num(); i++)
	{
		AActor *Actor = Level->Actors(i); 
		if (Actor && Actor->bHiddenEd && Actor->IsA(AInventory::StaticClass()) )
		{
			Pairs.AddItem( ActorIdxPair(Actor,i) );
			Level->Actors(i) = NULL;
		}
	}
	
	definePaths(ownerLevel);


	for (i=Pairs.Num()-1; i>=0; i--)
		if ( Level->Actors( Pairs(i).Idx) == NULL )
		{
			Level->Actors( Pairs(i).Idx) = Pairs(i).Actor;
			Pairs.Remove( i);
		}

	for ( i=0 ; i<Pairs.Num() ; i++ )
	{
		Level->Actors.AddItem( Pairs(i).Actor );
		debugf(TEXT("WTF"));
	}
	
	//Compact actor list?
	
	unguard;*/


	Level = ownerLevel;
	if ( !InventorySpotClass )
		InventorySpotClass = FindObjectChecked<UClass>( ANY_PACKAGE, TEXT("InventorySpot") );
	getScout();
	Level->GetLevelInfo()->NavigationPointList = NULL;

	// Add WarpZoneMarkers and InventorySpots
	debugf( NAME_DevPath, TEXT("Add WarpZone and Inventory markers") );
	for (INT i=0; i<Level->Actors.Num(); i++)
	{
		AActor *Actor = Level->Actors(i); 
		if ( Actor )
		{
			if ( Actor->IsA(AWarpZoneInfo::StaticClass()) )
			{
				if ( !findScoutStart(Actor->Location) || (Scout->Region.Zone != Actor->Region.Zone) )
				{
					Scout->SetCollisionSize(20, Scout->CollisionHeight);
					if ( !findScoutStart(Actor->Location) || (Scout->Region.Zone != Actor->Region.Zone) )
						Level->FarMoveActor(Scout, Actor->Location, 1, 1);
					Scout->SetCollisionSize(MINCOMMONRADIUS, Scout->CollisionHeight);
				}
				UClass* pathClass = FindObjectChecked<UClass>( ANY_PACKAGE, TEXT("WarpZoneMarker") );
				AWarpZoneMarker *newMarker = (AWarpZoneMarker *)Level->SpawnActor(pathClass, NAME_None, NULL, NULL, Scout->Location);
				newMarker->markedWarpZone = (AWarpZoneInfo *)Actor;
			}
			else if ( Actor->IsA(AInventory::StaticClass()) )
			{
				if ( !findScoutStart(Actor->Location) || (Abs(Scout->Location.Z - Actor->Location.Z) > Scout->CollisionHeight) )
					Level->FarMoveActor(Scout, Actor->Location + FVector(0,0,40 - Actor->CollisionHeight), 1, 1);

				AInventorySpot *newMarker = NULL;
				if ( !Actor->bHiddenEd && !((AInventory*)Actor)->myMarker  )
					newMarker = (AInventorySpot *)Level->SpawnActor( InventorySpotClass, NAME_None, NULL, NULL, Scout->Location);
				((AInventory *)Actor)->myMarker = newMarker;
				if ( newMarker )
					newMarker->markedItem = (AInventory *)Actor;
			}
		}
	}

	//calculate and add reachspecs to pathnodes
	debugf(NAME_DevPath,TEXT("Add reachspecs"));
	for (i=0; i<Level->Actors.Num(); i++)
	{
		AActor *Actor = Level->Actors(i); 
		if (Actor && Actor->IsA(ANavigationPoint::StaticClass()))
		{
			((ANavigationPoint *)Actor)->nextNavigationPoint = Level->GetLevelInfo()->NavigationPointList;
			Level->GetLevelInfo()->NavigationPointList = (ANavigationPoint *)Actor;
			addReachSpecs(Actor);
			debugf( NAME_DevPath, TEXT("Added reachspecs to %s"),Actor->GetName() );
		}
	}

	debugf(NAME_DevPath,TEXT("Added %d reachspecs"), Level->ReachSpecs.Num()); 
	//remove extra reachspecs from teleporters


	//prune excess reachspecs
	debugf(NAME_DevPath,TEXT("Prune reachspecs"));
	ANavigationPoint *Nav = Level->GetLevelInfo()->NavigationPointList;
	int numPruned = 0;
	while (Nav)
	{
		numPruned += Prune(Nav);
		Nav = Nav->nextNavigationPoint;
	}
	debugf(NAME_DevPath,TEXT("Pruned %d reachspecs"), numPruned);

	// Generate VisNoReach list
	for (ANavigationPoint *Path=Level->GetLevelInfo()->NavigationPointList;
		Path!=NULL;
		Path=Path->nextNavigationPoint)
	{
		addVisNoReach(Path);
	}

	Level->DestroyActor(Scout);
	debugf(NAME_DevPath,TEXT("All done"));
	unguard;
}


void FXC_PathBuilder::undefinePaths_XC (ULevel *ownerLevel)
{
	guard(FXC_PathBuilder::undefinePaths);
	Level = ownerLevel;

	//remove all reachspecs
	debugf(NAME_DevPath,TEXT("Remove %d old reachspecs"), Level->ReachSpecs.Num());
	Level->ReachSpecs.Empty();

	// clear navigationpointlist
	Level->GetLevelInfo()->NavigationPointList = NULL;

	//clear pathnodes
	for (INT i=0; i<Level->Actors.Num(); i++)
	{
		AActor *Actor = Level->Actors(i); 
		if (Actor && Actor->IsA(ANavigationPoint::StaticClass()))
		{
			if ( Actor->IsA(AInventorySpot::StaticClass()) && Actor->bHiddenEd ) //Only remove hidden inventory-spots
			{
				if ( ((AInventorySpot *)Actor)->markedItem )
					((AInventorySpot *)Actor)->markedItem->myMarker = NULL;
				Level->DestroyActor(Actor);
			}
			else if ( Actor->IsA(AWarpZoneMarker::StaticClass()) || Actor->IsA(ATriggerMarker::StaticClass()) || Actor->IsA(AButtonMarker::StaticClass()) )
				Level->DestroyActor(Actor);
			else
			{
				((ANavigationPoint *)Actor)->nextNavigationPoint = NULL;
				((ANavigationPoint *)Actor)->nextOrdered = NULL;
				((ANavigationPoint *)Actor)->prevOrdered = NULL;
				((ANavigationPoint *)Actor)->startPath = NULL;
				((ANavigationPoint *)Actor)->previousPath = NULL;
				for (INT i=0; i<16; i++)
				{
					((ANavigationPoint *)Actor)->Paths[i] = -1;
					((ANavigationPoint *)Actor)->upstreamPaths[i] = -1;
					((ANavigationPoint *)Actor)->PrunedPaths[i] = -1;
					((ANavigationPoint *)Actor)->VisNoReachPaths[i] = NULL;
				}
			}
		}
	}
	unguard;
}

void FXC_PathBuilder::addVisNoReach(AActor *start)
{
	guard(FXC_PathBuilder::addVisNoReach);

	ANavigationPoint *node = (ANavigationPoint *)start;
	//debugf("Add Reachspecs for node at (%f, %f, %f)", node->Location.X,node->Location.Y,node->Location.Z);

	if ( node->IsA(ALiftCenter::StaticClass()) )
		return;

	Scout->SetCollisionSize( HUMANRADIUS, HUMANHEIGHT );
	Level->FarMoveActor(Scout, node->Location, 1);
	Scout->MoveTarget = node;
	Scout->bCanDoSpecial = 1;
	AActor* newPath;
	INT i = 0;

	//debugf("VisNoReach for %s", node->GetName());
	for (ANavigationPoint *Path=Level->GetLevelInfo()->NavigationPointList;
		Path!=NULL;
		Path=Path->nextNavigationPoint)
	{
		FLOAT distSq = (node->Location - Path->Location).SizeSquared(); 
		if ( !Path->IsA(ALiftCenter::StaticClass()) && (Path != node) 
			&& (distSq < 4000000) && (i < 16) ) //FIXME - Pick right value!!!
		{
			FCheckResult Hit(1.0);
			Level->SingleLineCheck(Hit, Scout, Path->Location, node->Location, TRACE_VisBlocking);
			if ( !Hit.Actor )
			{
				FLOAT pathDist;
				if ( Scout->findPathToward(Path, 0, newPath, 1) )
					pathDist = ((ANavigationPoint *)newPath)->visitedWeight;
				else
				{
					pathDist = 200000000;
					//debugf(TEXT("NO PATH from %s to %s"), node->GetName(), Path->GetName());
				}
				//debugf("Path cost to %s = %f vs. dist %f",Path->GetName(), pathDist, distSq);
				if ( (pathDist != 10000000)
					&& (pathDist * pathDist > 4 * distSq) )
				{
					//debugf("Add %s to %s", Path->GetName(), node->GetName());
					node->VisNoReachPaths[i] = Path;
					i++;
				}
			}
		}
	}
	unguard;
}

void FXC_PathBuilder::addReachSpecs(AActor *start)
{
	guard(FXC_PathBuilder::addReachspecs);

	FReachSpec newSpec;
	ANavigationPoint *node = (ANavigationPoint *)start;
	//debugf("Add Reachspecs for node at (%f, %f, %f)", node->Location.X,node->Location.Y,node->Location.Z);

	if ( *(INT*)&node->Event != 0 ) //NAME_None
	{
		for (INT i=0; i<Level->Actors.Num(); i++)
		{
			AActor *Actor = Cast<ANavigationPoint>(Level->Actors(i));
			if ( Actor && ((Actor->GetFName() == node->Event) || (Actor->Tag == node->Event)) )
			{
				newSpec.Init();
				newSpec.CollisionRadius = 60;
				newSpec.CollisionHeight = 60;
				newSpec.reachFlags = R_SPECIAL;
				newSpec.Start = node;
				newSpec.End = Actor;
				newSpec.distance = 500;
				int pos = insertReachSpec(node->Paths, newSpec);
				if (pos != -1)
				{
					int iSpec = Level->ReachSpecs.AddItem(newSpec);
					node->Paths[pos] = iSpec;
					pos = insertReachSpec(((ANavigationPoint *)Actor)->upstreamPaths, newSpec);
					if (pos != -1)
						((ANavigationPoint *)Actor)->upstreamPaths[pos] = iSpec;
				}
			}
		}
	}

	if ( node->IsA(ALiftCenter::StaticClass()) )
	{
		FName myLiftTag = ((ALiftCenter *)node)->LiftTag;
		for (INT i=0; i<Level->Actors.Num(); i++)
		{
			AActor *Actor = Level->Actors(i); 
			if ( Actor && Actor->IsA(ALiftExit::StaticClass()) && (((ALiftExit *)Actor)->LiftTag == myLiftTag) ) 
			{
				newSpec.Init();
				newSpec.CollisionRadius = 60;
				newSpec.CollisionHeight = 60;
				newSpec.reachFlags = R_SPECIAL;
				newSpec.Start = node;
				newSpec.End = Actor;
				newSpec.distance = 500;
				int pos = insertReachSpec(node->Paths, newSpec);
				if (pos != -1)
				{
					int iSpec = Level->ReachSpecs.AddItem(newSpec);
					node->Paths[pos] = iSpec;
					pos = insertReachSpec(((ANavigationPoint *)Actor)->upstreamPaths, newSpec);
					if (pos != -1)
						((ANavigationPoint *)Actor)->upstreamPaths[pos] = iSpec;
				}
				newSpec.Init();
				newSpec.CollisionRadius = 60;
				newSpec.CollisionHeight = 60;
				newSpec.reachFlags = R_SPECIAL;
				newSpec.Start = Actor;
				newSpec.End = node;
				newSpec.distance = 500;
				pos = insertReachSpec(((ANavigationPoint *)Actor)->Paths, newSpec);
				if (pos != -1)
				{
					int iSpec = Level->ReachSpecs.AddItem(newSpec);
					((ANavigationPoint *)Actor)->Paths[pos] = iSpec;
					pos = insertReachSpec(node->upstreamPaths, newSpec);
					if (pos != -1)
						node->upstreamPaths[pos] = iSpec;
				}
			}
		}
		return;
	}

	if ( node->IsA(ATeleporter::StaticClass()) || node->IsA(AWarpZoneMarker::StaticClass()) )
	{
		for (INT i=0; i<Level->Actors.Num(); i++)
		{
			int bFoundMatch = 0;
			AActor *Actor = Level->Actors(i); 
			if ( node->IsA(ATeleporter::StaticClass()) )
			{
				if ( Actor && Actor->IsA(ATeleporter::StaticClass()) && (Actor != node) ) 
					bFoundMatch = (((ATeleporter *)node)->URL==*Actor->Tag);
			}
			else if ( Actor && Actor->IsA(AWarpZoneMarker::StaticClass()) && (Actor != node) )
				bFoundMatch = (((AWarpZoneMarker *)node)->markedWarpZone->OtherSideURL==*((AWarpZoneMarker *)Actor)->markedWarpZone->ThisTag);

			if ( bFoundMatch )
			{
				newSpec.Init();
				newSpec.CollisionRadius = 150;
				newSpec.CollisionHeight = 150;
				newSpec.reachFlags = R_SPECIAL;
				newSpec.Start = node;
				newSpec.End = Actor;
				newSpec.distance = 100;
				int pos = insertReachSpec(node->Paths, newSpec);
				if (pos != -1)
				{
					int iSpec = Level->ReachSpecs.AddItem(newSpec);
					//debugf("     Add teleport reachspec %d to node at (%f, %f, %f)", iSpec, Actor->Location.X,Actor->Location.Y,Actor->Location.Z);
					node->Paths[pos] = iSpec;
					pos = insertReachSpec(((ANavigationPoint *)Actor)->upstreamPaths, newSpec);
					if (pos != -1)
						((ANavigationPoint *)Actor)->upstreamPaths[pos] = iSpec;
				}
				break;
			}
		}
	}
	
	
	for (INT i=0; i<Level->Actors.Num(); i++)
	{
		AActor *Actor = Level->Actors(i); 
		if (Actor && Actor->IsA(ANavigationPoint::StaticClass()) && !Actor->IsA(ALiftCenter::StaticClass()) 
				&& (Actor != node) && ((node->Location - Actor->Location).SizeSquared() < ScanRangeSq)
				&& (!node->bOneWayPath || (((Actor->Location - node->Location) | node->Rotation.Vector()) > 0)) )
		{
			newSpec.Init();
			if (newSpec.defineFor(node, Actor, Scout))
			{
				int pos = insertReachSpec(node->Paths, newSpec);
				if (pos != -1)
				{
					int iSpec = Level->ReachSpecs.AddItem(newSpec);
					//debugf("     Add reachspec %d to node at (%f, %f, %f)", iSpec, Actor->Location.X,Actor->Location.Y,Actor->Location.Z);
					node->Paths[pos] = iSpec;
					pos = insertReachSpec(((ANavigationPoint *)Actor)->upstreamPaths, newSpec);
					if (pos != -1)
						((ANavigationPoint *)Actor)->upstreamPaths[pos] = iSpec;
				} 
			}
		}
	}
	unguard;
}

int FXC_PathBuilder::findScoutStart(FVector start)
{
	guard(FXC_PathBuilder::findScoutStart);
	
	if (Level->FarMoveActor(Scout, start)) //move Scout to starting point
	{
		//slide down to floor
		FCheckResult Hit(1.0);
		FVector Down = FVector(0,0, -50);
		Hit.Normal.Z = 0.0;
		INT iters = 0;
		while (Hit.Normal.Z < 0.7)
		{
			Level->MoveActor(Scout, Down, Scout->Rotation, Hit, 1,1);
			if ((Hit.Time < 1.0) && (Hit.Normal.Z < 0.7)) 
			{
				//adjust and try again
				FVector OldHitNormal = Hit.Normal;
				FVector Delta = (Down - Hit.Normal * (Down | Hit.Normal)) * (1.0 - Hit.Time);
				if( (Delta | Down) >= 0 )
				{
					Level->MoveActor(Scout, Delta, Scout->Rotation, Hit, 1,1);
					if ((Hit.Time < 1.0) && (Hit.Normal.Z < 0.7))
					{
						FVector downDir = Down.SafeNormal();
						Scout->TwoWallAdjust(downDir, Delta, Hit.Normal, OldHitNormal, Hit.Time);
						Level->MoveActor(Scout, Delta, Scout->Rotation, Hit, 1,1);
					}
				}
			}
			iters++;
			if (iters >= 50)
			{
				debugf(NAME_DevPath,TEXT("No valid start found"));
				return 0;
			}
		}
		//debugf(NAME_DevPath,"scout placed on valid floor");
		return 1;
 	}

	debugf(NAME_DevPath,TEXT("Scout didn't fit"));
	return 0;
	unguard;
}

int FXC_PathBuilder::Prune(AActor *Node)
{
	guard(FPathBuilder::Prune);
	
	int pruned = 0;
	ANavigationPoint* NavNode = (ANavigationPoint *)Node;

	int n,j;
	FReachSpec alternatePath, straightPath, part1, part2;
	int i=0;
	guard(BigLoop);
	while ( (i<16) && (NavNode->upstreamPaths[i] != -1) )
	{
		guard(Access_1);
		part1 = Level->ReachSpecs(NavNode->upstreamPaths[i]);
		unguard;
		guard(MidLoop);
		n=0;
		while ( (n<16) && (NavNode->Paths[n] != -1) )
		{
			guard(Access_2);
			part2 = Level->ReachSpecs(NavNode->Paths[n]);
			unguard;
			INT straightPathIndex = specFor(part1.Start, part2.End);
			guard(Straight);
			if (straightPathIndex != -1)
			{
				straightPath = Level->ReachSpecs(straightPathIndex);
				alternatePath = part1 + part2;
				if ( ((float)straightPath.distance * 1.2 >= (float)alternatePath.distance) 
					&& ((alternatePath <= straightPath) || straightPath.BotOnlyPath() 
						|| alternatePath.MonsterPath()) )
				{
					//prune straightpath
					pruned++;
					j=0;
					ANavigationPoint* StartNode = (ANavigationPoint *)(straightPath.Start);
					ANavigationPoint* EndNode = (ANavigationPoint *)(straightPath.End);
					//debugf("Prune reachspec %d from %s to %s because of %s", straightPathIndex,
					//	StartNode->GetName(), EndNode->GetName(), NavNode->GetName()); 
					while ( (j<15) && (StartNode->Paths[j] != straightPathIndex) )
						j++;
					if ( StartNode->Paths[j] == straightPathIndex )
					{
						while ( (j<15) && (StartNode->Paths[j] != -1) )
						{
							StartNode->Paths[j] = StartNode->Paths[j+1];
							j++;
						}
						StartNode->Paths[15] = -1;
					}

					j=0;
					while ( (j<15) && (StartNode->PrunedPaths[j] != -1) )
						j++;
					StartNode->PrunedPaths[j] = straightPathIndex;
					Level->ReachSpecs(straightPathIndex).bPruned = 1;
					guard(EndHere);
					j=0;
					while ( (j<15) && (EndNode->upstreamPaths[j] != straightPathIndex) )
						j++;
					if ( EndNode->upstreamPaths[j] == straightPathIndex )
					{
						while ( (j<15) && (EndNode->upstreamPaths[j] != -1) )
						{
							EndNode->upstreamPaths[j] = EndNode->upstreamPaths[j+1];
							j++;
						}
						EndNode->upstreamPaths[15] = -1;
					}
					unguard;
					// Note that all specs remain in reachspec list and still referenced in PrunedPaths[]
					// but removed from visible Navigation graph:
				}
			}
			unguard;
			n++;
		}
		unguard;
		i++;
	}
	unguard;

	return pruned;
	unguard;
}

void FXC_PathBuilder::getScout()
{
	guard(FPathBuilder::getScout);
	Scout = NULL;
	for( INT i=0; i<Level->Actors.Num(); i++ )
	{
		AActor *Actor = Level->Actors(i); 
		if (Actor && Actor->IsA(AScout::StaticClass()))
			Scout = (APawn *)Actor;
	}
	if( !Scout )
	{
		UClass *scoutClass = FindObjectChecked<UClass>( ANY_PACKAGE, TEXT("Scout") );
		Scout = (APawn *)Level->SpawnActor( scoutClass );
	}
	Scout->SetCollision(1,1,1);
	Scout->GroundSpeed = 380;
	Scout->bCollideWorld = 1;
	Level->SetActorZone( Scout,1,1 );
	return;
	unguard;
}

int FXC_PathBuilder::specFor(AActor* Start, AActor* End)
{
	guard(FPathBuilder::specFor);
	ANavigationPoint* StartNode = Cast<ANavigationPoint>(Start);
	if ( !StartNode )
		return -1;
	for ( int i=0 ; i<16 && StartNode->Paths[i] != -1 ; i++ )
		if ( Level->ReachSpecs(StartNode->Paths[i]).End == End )
			return StartNode->Paths[i];
	return -1;
	unguard;
}

int FXC_PathBuilder::insertReachSpec(INT *SpecArray, FReachSpec &Spec)
{
	guard(FPathBuilder::insertReachSpec);

	int n = 0;
	while ( (n < 16) && (SpecArray[n] != -1) && (Level->ReachSpecs(SpecArray[n]).distance > Spec.distance) )
		n++;
	int pos = n;

	if (SpecArray[15] == -1) //then not full
	{
		if (SpecArray[n] != -1)
		{
			int current = SpecArray[n];
			while (n < 15)
			{
				int next = SpecArray[n+1];
				SpecArray[n+1] = current;
				current = next;
				if (next == -1)
					n = 15;
				n++;
			}
		}
	}
	else if (n == 0) // current is bigger than biggest
	{
		//debugf("Node out of Reachspecs - don't add!");
		return -1;
	}
	else //full, so remove from front
	{
		//debugf("Node out of Reachspecs -add at %d !", pos-1);

		n--;
		pos = n;
		int current = SpecArray[n];
		while (n > 0)
		{
			int next = SpecArray[n-1];
			SpecArray[n-1] = current;
			current = next;
			n--;
		}
	}
	return pos;

	unguard;
}